// src/index.ts
function generateJSONPatch(before, after, config = {}) {
  const { comparator, propertyFilter } = config;
  const patch = [];
  const hasPropertyFilter = typeof propertyFilter === "function";
  function compareArrayByIndex(leftArr, rightArr, path) {
    let currentIndex = 0;
    const maxLength = Math.max(leftArr.length, rightArr.length);
    for (let i = 0; i < maxLength; i++) {
      const newPathIndex = `${path}/${currentIndex++}`;
      if (i < leftArr.length && i < rightArr.length) {
        compareObjects(newPathIndex, leftArr[i], rightArr[i]);
      } else if (i >= leftArr.length && i < rightArr.length) {
        patch.push({ op: "add", path: newPathIndex, value: rightArr[i] });
      } else if (i < leftArr.length && i >= rightArr.length) {
        patch.push({ op: "remove", path: newPathIndex });
        currentIndex--;
      }
    }
  }
  function compareArrayByHash(leftArr, rightArr, path) {
    var _a;
    if (typeof comparator !== "function") {
      throw Error("No comparator function provided");
    }
    const leftHashes = leftArr.map((value) => comparator(value, { side: "left", path }));
    const rightHashes = rightArr.map((value) => comparator(value, { side: "right", path }));
    let currentIndex = 0;
    const targetHashes = [];
    for (let i = 0; i < leftHashes.length; i++) {
      const newPathIndex = `${path}/${currentIndex++}`;
      const rightHashIndex = rightHashes.indexOf(leftHashes[i]);
      if (rightHashIndex >= 0) {
        compareObjects(newPathIndex, leftArr[i], rightArr[rightHashIndex]);
        targetHashes.push(leftHashes[i]);
      } else {
        patch.push({ op: "remove", path: newPathIndex });
        currentIndex--;
      }
    }
    const toBeAddedHashes = rightHashes.filter((hash) => !targetHashes.includes(hash));
    for (const toBeAddedHash of toBeAddedHashes) {
      patch.push({
        op: "add",
        path: `${path}/${currentIndex++}`,
        value: rightArr[rightHashes.indexOf(toBeAddedHash)]
      });
      targetHashes.push(toBeAddedHash);
    }
    if ((_a = config.array) == null ? void 0 : _a.ignoreMove) {
      return;
    }
    for (let i = rightHashes.length - 1; i >= 0; i--) {
      const hash = rightHashes[i];
      const targetIndex = rightHashes.indexOf(hash);
      const currentIndex2 = targetHashes.indexOf(hash);
      if (currentIndex2 !== targetIndex) {
        patch.push({ op: "move", from: `${path}/${currentIndex2}`, path: `${path}/${targetIndex}` });
        moveArrayElement(targetHashes, currentIndex2, targetIndex);
      }
    }
  }
  function compareArrays(leftArr, rightArr, path) {
    if (JSON.stringify(leftArr) === JSON.stringify(rightArr))
      return;
    if (comparator) {
      compareArrayByHash(leftArr, rightArr, path);
    } else {
      compareArrayByIndex(leftArr, rightArr, path);
    }
  }
  function compareObjects(path, leftObj, rightObj) {
    const isArrayAtTop = path === "" && (Array.isArray(leftObj) && Array.isArray(rightObj));
    if (isPrimitiveValue(leftObj) || isPrimitiveValue(rightObj)) {
      if (leftObj !== rightObj) {
        patch.push({ op: "replace", path, value: rightObj });
      }
      return;
    }
    if (isArrayAtTop) {
      return compareArrays(leftObj, rightObj, "");
    }
    if (Array.isArray(leftObj) && !Array.isArray(rightObj) || !Array.isArray(leftObj) && Array.isArray(rightObj)) {
      patch.push({ op: "replace", path, value: rightObj });
      return;
    }
    for (const rightKey in rightObj) {
      if (hasPropertyFilter && !propertyFilter(rightKey, { side: "right", path }))
        continue;
      let newPath = isArrayAtTop && path === "" ? `/${rightKey}` : `${path}/${rightKey}`;
      const leftValue = leftObj[rightKey];
      const rightValue = rightObj[rightKey];
      if (Array.isArray(leftValue) && Array.isArray(rightValue)) {
        compareArrays(leftValue, rightValue, newPath);
      } else if (isJsonObject(rightValue)) {
        if (isJsonObject(leftValue)) {
          compareObjects(newPath, leftValue, rightValue);
        } else {
          patch.push({ op: "replace", path: newPath, value: rightValue });
        }
      } else if (!leftObj.hasOwnProperty(rightKey)) {
        patch.push({ op: "add", path: newPath, value: rightValue });
      } else if (leftValue !== rightValue) {
        patch.push({ op: "replace", path: newPath, value: rightValue });
      }
    }
    for (const leftKey in leftObj) {
      if (!leftObj.hasOwnProperty(leftKey) || hasPropertyFilter && !propertyFilter(leftKey, { side: "left", path }))
        continue;
      if (!rightObj.hasOwnProperty(leftKey)) {
        let newPath = isArrayAtTop && path === "" ? `/${leftKey}` : `${path}/${leftKey}`;
        patch.push({ op: "remove", path: newPath });
      }
    }
  }
  compareObjects("", before, after);
  return [...patch];
}
function isPrimitiveValue(value) {
  return value === void 0 || value === null || typeof value === "undefined" || typeof value === "string" || typeof value === "number" || typeof value === "boolean";
}
function isJsonObject(value) {
  return (value == null ? void 0 : value.constructor) === Object;
}
function moveArrayElement(array, from, to) {
  array.splice(to, 0, array.splice(from, 1)[0]);
}
function pathInfo(path) {
  const segments = path.split("/");
  const length = segments.length;
  const last = segments[length - 1];
  return { segments, length, last };
}
export {
  generateJSONPatch,
  pathInfo
};
